--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 05.05.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 05.05.2020 - Release version

Important:
No changes are to be made to this script without permission from GtX | Andy
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


local loadError
local popupMessage

local debugActive = false
local versionString = "1.0.0.0"

local modName = g_currentModName
local modDirectory = g_currentModDirectory
local specializationName = string.format("%s.multiplayerVehicleKey", modName)

local function getIsMultiplayerGame()
    if g_dedicatedServerInfo ~= nil then
        return true
    end

    if g_careerScreen ~= nil then
        return g_careerScreen.isMultiplayer or debugActive
    end

    return true
end

local function onStartMission(mission)
    if g_dedicatedServerInfo == nil then
        if mission.currentVehicleKey == nil then
            if g_multiplayerVehicleKeys ~= nil then
                mission.currentVehicleKey = g_multiplayerVehicleKeys[mission.playerUserId]
            end
        end

        if mission.currentVehicleKey ~= nil then
            local vehicleName

            if mission.currentVehicleKey.getFullName ~= nil then
                vehicleName = mission.currentVehicleKey:getFullName()
            end

            if vehicleName ~= nil then
                mission.hud:addSideNotification({0.07, 0.827, 0.898, 1}, string.format(g_i18n:getText("info_holdingKeys", modName), tostring(vehicleName)), 8000)
            end
        end
    end
end

local function finalizeVehicleTypes(vehicleTypeManager)
    local numInserted = 0

    if specializationName ~= nil then
        local specializationObject = g_specializationManager:getSpecializationObjectByName(specializationName)

        if specializationObject ~= nil then
            for typeName, typeEntry in pairs(vehicleTypeManager.vehicleTypes) do
                if specializationObject.prerequisitesPresent(typeEntry.specializations) then
                    g_vehicleTypeManager:addSpecialization(typeName, specializationName)
                    numInserted = numInserted + 1
                end
            end
        end
    end

    if debugActive == true then
        print(string.format("  Debug:  Inserted specialisation 'multiplayerVehicleKey' into %d vehicle types.", numInserted))
    end
end

local function addPopupMessage(message)
    popupMessage = {
        delay = 1500,
        modName = modName,
        message = message,

        update = function(self, dt)
            self.delay = self.delay - dt

            if self.delay < 0 then
                if not g_gui:getIsGuiVisible() then
                    if g_currentMission.hud ~= nil then
                        g_currentMission.hud:showInGameMessage("", self.message or "", -1, nil, nil, nil)
                    end

                    removeModEventListener(self)
                    popupMessage = nil
                end
            end
        end
    }

    addModEventListener(popupMessage)
end

local function addLoadError(name)
    loadError = {
        name = name,
        modName = modName,
        startUpdateTime = 1800,
        versionString = versionString,

        update = function(self, dt)
            self.startUpdateTime = self.startUpdateTime - dt

            if self.startUpdateTime < 0 then
                local title = string.format("Multiplayer Vehicle Keys - Version %s", self.versionString)
                local text = string.format(g_i18n:getText("MVK_loadError", self.modName), self.modName, self.name)

                if g_dedicatedServerInfo == nil then
                    if not g_gui:getIsGuiVisible() then
                        g_gui:showYesNoDialog({
                            title = title,
                            text = text,
                            dialogType = DialogElement.TYPE_LOADING,
                            callback = self.openModHubLink,
                            target = nil,
                            yesText = g_i18n:getText("button_ok"),
                            noText = g_i18n:getText("button_modHubDownload")
                        })
                    end
                else
                    print("\n" .. text .. "\n    - https://farming-simulator.com/mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0" .. "\n")
                    self.openModHubLink(true)
                end
            end
        end,

        openModHubLink = function(ignore)
            if ignore == false then
                local language = g_languageShort
                local link = "mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                if language == "de" or language == "fr" then
                    link = "mods.php?lang=" .. language .. "&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                end

                openWebFile(link, "")
            end

            removeModEventListener(loadError)
            loadError = nil
        end
    }

    addModEventListener(loadError)
end

local function validate(name)
    local mod = g_modManager:getModByName(name)
    versionString = tostring(mod.version) or versionString

    if mod.modName == "FS19_MultiplayerVehicleKeys" or
        mod.modName == "FS19_MultiplayerVehicleKeys_update" then
        return mod.author:len() == 3, mod.author
    end

    return false, mod.author
end

function delete()
    getfenv(0)["g_multiplayerVehicleKeys"] = nil
end

local function onUserAdded(user)
    if user:getId() == g_currentMission.playerUserId then
        if g_currentMission.enterables ~= nil then
            for _, vehicle in pairs (g_currentMission.enterables) do
                if vehicle.getIsKeyOwnerFromUniqueUserId ~= nil and vehicle:getIsKeyOwnerFromUniqueUserId(user:getUniqueUserId()) then
                    vehicle:addKeyOwner(user, nil, false)
                    break
                end
            end
        end
    end
end

local function onUserRemoved(user)
    if g_server ~= nil and g_currentMission ~= nil then
        if g_currentMission.enterables ~= nil then
            for _, vehicle in pairs (g_currentMission.enterables) do
                if vehicle.getIsKeyOwnerFromUniqueUserId ~= nil and vehicle:getIsKeyOwnerFromUniqueUserId(user:getUniqueUserId()) then
                    vehicle:removeKeyOwner(true, user.connection, false)
                    break
                end
            end
        end
    end
end

local function notifyPlayerFarmChanged(player)
    if player == g_currentMission.player then
        if g_currentMission.currentVehicleKey ~= nil then
            g_currentMission.currentVehicleKey:removeKeyOwner(false, false)
        end
    end
end

local function init()
    if g_multiplayerVehicleKeys ~= nil then
        return
    end

    local isValid, author = validate(modName)
    if isValid then
        if getIsMultiplayerGame() then
            getfenv(0)["g_multiplayerVehicleKeys"] = {}

            source(modDirectory .. "scripts/specializations/events/SetCurrentKeyOwnerEvent.lua")
            g_specializationManager:addSpecialization("multiplayerVehicleKey", "MultiplayerVehicleKey", modDirectory .. "scripts/specializations/MultiplayerVehicleKey.lua")

            FSBaseMission.delete = Utils.appendedFunction(FSBaseMission.delete, delete)
            Mission00.onStartMission = Utils.appendedFunction(Mission00.onStartMission, onStartMission)
            VehicleTypeManager.finalizeVehicleTypes = Utils.prependedFunction(VehicleTypeManager.finalizeVehicleTypes, finalizeVehicleTypes)

            g_messageCenter:subscribe(MessageType.USER_ADDED, onUserAdded, nil, nil)
            g_messageCenter:subscribe(MessageType.PLAYER_FARM_CHANGED, notifyPlayerFarmChanged, nil, nil)

            if g_server ~= nil then
                g_messageCenter:subscribe(MessageType.USER_REMOVED, onUserRemoved, nil, nil)
            end
        else
            addPopupMessage(string.format(g_i18n:getText("warning_multiplayerNotDetected", modName), modName))
        end
    else
        addLoadError(author)
    end
end

init()
