--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 05.05.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Mod idea / concept:
Johnny Vee and GtX

History:
V 1.0.0.0 @ 05.05.2020 - Release version

Important:
No changes are to be made to this script without permission from GtX | Andy
Use of this specialization within other mods is strictly prohibited!

An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
Die Verwendung dieser Spezialisierung innerhalb anderer Mods ist strengstens untersagt!
]]


MultiplayerVehicleKey = {}

MultiplayerVehicleKey.MOD_NAME = g_currentModName

MultiplayerVehicleKey.NOTIFICATION_TIME = 4000

MultiplayerVehicleKey.INGAME_NOTIFICATION_KEYS = {
    0.07,
    0.827,
    0.898,
    1
}

function MultiplayerVehicleKey.prerequisitesPresent(specializations)
    if g_multiplayerVehicleKeys ~= nil then
        return not SpecializationUtil.hasSpecialization(Locomotive, specializations) and
            SpecializationUtil.hasSpecialization(Motorized, specializations) and
            SpecializationUtil.hasSpecialization(Enterable, specializations)
    end

    return false
end

function MultiplayerVehicleKey.registerEvents(vehicleType)
    SpecializationUtil.registerEvent(vehicleType, "onAddKeyOwner")
    SpecializationUtil.registerEvent(vehicleType, "onRemoveKeyOwner")
end

function MultiplayerVehicleKey.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "addKeyOwner",                   MultiplayerVehicleKey.addKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "removeKeyOwner",                MultiplayerVehicleKey.removeKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "setKeyOwnerData",               MultiplayerVehicleKey.setKeyOwnerData)
    SpecializationUtil.registerFunction(vehicleType, "getHasKeyOwner",                MultiplayerVehicleKey.getHasKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "canMasterUserAccess",           MultiplayerVehicleKey.canMasterUserAccess)
    SpecializationUtil.registerFunction(vehicleType, "getKeyOwnerName",               MultiplayerVehicleKey.getKeyOwnerName)
    SpecializationUtil.registerFunction(vehicleType, "getCurrentKeyOwnerId",          MultiplayerVehicleKey.getCurrentKeyOwnerId)
    SpecializationUtil.registerFunction(vehicleType, "getUserCurrentVehicleKey",      MultiplayerVehicleKey.getUserCurrentVehicleKey)
    SpecializationUtil.registerFunction(vehicleType, "getIsKeyOwnerFromUniqueUserId", MultiplayerVehicleKey.getIsKeyOwnerFromUniqueUserId)
    SpecializationUtil.registerFunction(vehicleType, "getIsVehicleKeyAvailable",      MultiplayerVehicleKey.getIsVehicleKeyAvailable)
end

function MultiplayerVehicleKey.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "enterVehicle",              MultiplayerVehicleKey.enterVehicle)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setIsTabbable",             MultiplayerVehicleKey.setIsTabbable)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanMotorRun",            MultiplayerVehicleKey.getCanMotorRun)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getMotorNotAllowedWarning", MultiplayerVehicleKey.getMotorNotAllowedWarning)
end

function MultiplayerVehicleKey.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad",                 MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished",         MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onPreDelete",            MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream",           MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream",          MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick",           MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onLeaveVehicle",         MultiplayerVehicleKey)
    SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", MultiplayerVehicleKey)
end

function MultiplayerVehicleKey:onLoad(savegame)
    self.spec_enterableMultiplayerKey = self[string.format("spec_%s.multiplayerVehicleKey", MultiplayerVehicleKey.MOD_NAME)]

    local spec = self.spec_enterableMultiplayerKey

    spec.keyOwner = nil
    spec.keyOwnerName = ""

    spec.currentUser = nil
    spec.masterUserCanAccess = false

    spec.isTabbable = self:getIsTabbable()

    spec.keyOwnerInfo = g_i18n:getText("info_keyOwner", self.customEnvironment)
    spec.holdingKeysInfo = g_i18n:getText("info_holdingKeys", self.customEnvironment)
    spec.keysReturnedInfo = g_i18n:getText("info_keysReturned", self.customEnvironment)

    spec.actionRemoveKey = g_i18n:getText("action_removeVehicleKey", self.customEnvironment)
    spec.actionReturnKey = g_i18n:getText("action_returnVehicleKey", self.customEnvironment)

    spec.invalidUserWarning = string.format("%s\n%s", g_i18n:getText("warning_youDontHaveAccessToThis"), spec.keyOwnerInfo)
    spec.holdingKeysWarning = string.format("%s\n%s", g_i18n:getText("warning_actionNotAllowedNow"), spec.holdingKeysInfo)

    -- Fix as this is an integer on load and should be boolean.
    self.spec_motorized.showTurnOnMotorWarning = false
end

function MultiplayerVehicleKey:onLoadFinished(savegame)
    if savegame ~= nil then
        local key = string.format("%s.%s.multiplayerVehicleKey", savegame.key, MultiplayerVehicleKey.MOD_NAME)
        local uniqueUserId = getXMLString(savegame.xmlFile, key .. "#uniqueUserId")

        if uniqueUserId ~= nil then
            local nickname = Utils.getNoNil(getXMLString(savegame.xmlFile, key .. "#lastNickname"), "")
            local userId = 0

            if savegame.resetVehicles then
                userId = Utils.getNoNil(getXMLInt(savegame.xmlFile, key .. "#userId"), userId)
            else
                if getUniqueUserId() == uniqueUserId then
                    userId = g_currentMission.playerUserId
                end
            end

            if userId > 0 then
                self:addKeyOwner(nil, userId, true)
            else
                self:setKeyOwnerData(uniqueUserId, nickname, userId, false, false)
            end
        end
    end
end

function MultiplayerVehicleKey:onPreDelete()
    self:removeKeyOwner(false, nil, true)
end

function MultiplayerVehicleKey:saveToXMLFile(xmlFile, key, usedModNames)
    local spec = self.spec_enterableMultiplayerKey

    if spec.keyOwner ~= nil and spec.keyOwner.uniqueUserId ~= nil then
        setXMLString(xmlFile, key .. "#uniqueUserId", spec.keyOwner.uniqueUserId)
        setXMLString(xmlFile, key .. "#lastNickname", spec.keyOwner.nickname or "")
        setXMLInt(xmlFile, key .. "#userId", spec.keyOwner.id or 0)
    end
end

function MultiplayerVehicleKey:onReadStream(streamId, connection)
    local spec = self.spec_enterableMultiplayerKey

    if streamReadBool(streamId) then
        local uniqueUserId = streamReadString(streamId)
        local nickname = streamReadString(streamId)
        local userId = streamReadInt8(streamId)

        spec.isTabbable = streamReadBool(streamId)

        if userId > 0 and g_currentMission ~= nil and g_currentMission.gameStarted then
            self:addKeyOwner(nil, userId, true)
        else
            self:setKeyOwnerData(uniqueUserId, nickname, userId, userId > 0, false)
        end
    end
end

function MultiplayerVehicleKey:onWriteStream(streamId, connection)
    local spec = self.spec_enterableMultiplayerKey

    if streamWriteBool(streamId, spec.keyOwner ~= nil) then
        streamWriteString(streamId, spec.keyOwner.uniqueUserId)
        streamWriteString(streamId, spec.keyOwner.nickname)
        streamWriteInt8(streamId, spec.keyOwner.id or 0)

        streamWriteBool(streamId, spec.isTabbable)
    end
end

function MultiplayerVehicleKey:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    if self.isClient then
        if isActiveForInputIgnoreSelection then
            local spec = self.spec_enterableMultiplayerKey
            local actionEvent = spec.actionEvents[InputAction.VEHICLE_KEY_INTERACT]

            if actionEvent ~= nil then
                if self:getIsMotorStarted() then
                    g_inputBinding:setActionEventTextPriority(actionEvent.actionEventId, GS_PRIO_LOW)
                else
                    g_inputBinding:setActionEventTextPriority(actionEvent.actionEventId, GS_PRIO_VERY_HIGH)
                end

                g_inputBinding:setActionEventText(actionEvent.actionEventId, MultiplayerVehicleKey.getActionEventText(spec))
            end
        end
    end
end

function MultiplayerVehicleKey:onLeaveVehicle()
    self.spec_enterableMultiplayerKey.currentUser = nil
end

function MultiplayerVehicleKey:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    local spec = self.spec_enterableMultiplayerKey

    if isActiveForInputIgnoreSelection then
        self:clearActionEventsTable(spec.actionEvents)

        if self:getIsVehicleKeyAvailable(true) then
            if self:getIsActiveForInput(true, true) then
                local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.VEHICLE_KEY_INTERACT, self, MultiplayerVehicleKey.actionEventVehicleKeyInteract, false, true, false, true, nil)
                g_inputBinding:setActionEventText(actionEventId, MultiplayerVehicleKey.getActionEventText(spec))
            end
        end
    end
end

function MultiplayerVehicleKey:actionEventVehicleKeyInteract(actionName, inputValue, callbackState, isAnalog)
    if self:getIsEntered() then
        local currentKeys = self:getUserCurrentVehicleKey(userId)

        if currentKeys ~= nil then
            if currentKeys == self then
                self:removeKeyOwner(false, nil, false)
            else
                if currentKeys.getFullName ~= nil then
                    g_currentMission:showBlinkingWarning(string.format(self.spec_enterableMultiplayerKey.holdingKeysWarning, tostring(currentKeys:getFullName())), MultiplayerVehicleKey.NOTIFICATION_TIME)
                end
            end
        else
            self:addKeyOwner(nil, g_currentMission.playerUserId, false)
        end
    end
end

function MultiplayerVehicleKey:getIsVehicleKeyAvailable(inputCheck)
    local spec = self.spec_enterableMultiplayerKey

    if spec.keyOwner == nil then
        return true
    end

    if not inputCheck and self:getIsAIActive() then
        return true
    end

    if spec.currentUser ~= nil then
        if spec.currentUser.uniqueUserId == spec.keyOwner.uniqueUserId then
            return true
        end

        return not inputCheck and spec.currentUser.isMasterUser and spec.masterUserCanAccess
    end

    return false
end

function MultiplayerVehicleKey:addKeyOwner(user, userId, noEventSend)
    local spec = self.spec_enterableMultiplayerKey

    if user == nil then
        if spec.currentUser ~= nil and spec.currentUser.id == userId then
            user = spec.currentUser
        else
            user = g_currentMission.userManager:getUserByUserId(userId)
        end
    end

    if user ~= nil then
        userId = userId or user.id

        if noEventSend == nil or noEventSend == false then
            if g_server ~= nil then
                g_server:broadcastEvent(SetCurrentKeyOwnerEvent:new(self, userId, false), nil, nil, self)
            else
                g_client:getServerConnection():sendEvent(SetCurrentKeyOwnerEvent:new(self, userId, false))
            end
        end

        local keyOwnerData, _ = self:setKeyOwnerData(user.uniqueUserId, user.nickname, userId, userId > 0, false)
        SpecializationUtil.raiseEvent(self, "onAddKeyOwner", userId, keyOwnerData)
    end
end

-- 'ignoreConnection' Do not want to send to the disconnecting client if the originated with the server. i.e onUserRemoved
function MultiplayerVehicleKey:removeKeyOwner(setOffline, ignoreConnection, noEventSend)
    local spec = self.spec_enterableMultiplayerKey

    if spec.keyOwner ~= nil then
        local uniqueUserId, nickname = nil, ""
        setOffline = Utils.getNoNil(setOffline, false)

        if noEventSend == nil or noEventSend == false then
            if g_server ~= nil then
                g_server:broadcastEvent(SetCurrentKeyOwnerEvent:new(self, nil, setOffline), nil, ignoreConnection, self)
            else
                g_client:getServerConnection():sendEvent(SetCurrentKeyOwnerEvent:new(self, nil, setOffline))
            end
        end

        if setOffline then
            uniqueUserId = spec.keyOwner.uniqueUserId
            nickname = spec.keyOwner.nickname
        end

        local _, userId = self:setKeyOwnerData(uniqueUserId, nickname, spec.keyOwner.id, false, setOffline)
        SpecializationUtil.raiseEvent(self, "onRemoveKeyOwner", userId, setOffline)
    end
end

function MultiplayerVehicleKey:setKeyOwnerData(uniqueUserId, nickname, userId, isOnline, setOffline)
    local spec = self.spec_enterableMultiplayerKey

    local isTabbable = false
    local validId = userId > 0

    if uniqueUserId ~= nil then
        spec.keyOwner = {
            uniqueUserId = uniqueUserId,
            nickname = nickname,
            id = userId,
            isOnline = isOnline
        }

        if validId and g_multiplayerVehicleKeys ~= nil then
            local oldVehicle = g_multiplayerVehicleKeys[userId]

            if oldVehicle ~= nil and oldVehicle ~= self then
                if oldVehicle.removeKeyOwner ~= nil then
                    oldVehicle:removeKeyOwner(false, nil, true)
                end
            end

            g_multiplayerVehicleKeys[userId] = self
        end
    else
        if validId and g_multiplayerVehicleKeys ~= nil then
            if g_multiplayerVehicleKeys[userId] == self then
                g_multiplayerVehicleKeys[userId] = nil
            end
        end

        isTabbable = true
        spec.keyOwner = nil
    end

    if validId and g_currentMission.playerUserId == userId then
        if isTabbable or setOffline then
            g_currentMission.currentVehicleKey = nil
            MultiplayerVehicleKey.addSideNotification(isTabbable, spec.keysReturnedInfo, self:getFullName())
        else
            isTabbable = true
            g_currentMission.currentVehicleKey = self
            MultiplayerVehicleKey.addSideNotification(isOnline, spec.holdingKeysInfo, self:getFullName())
        end
    end

    spec.keyOwnerName = nickname or ""
    spec.masterUserCanAccess = not isOnline

    self:setIsTabbable(spec.isTabbable and isTabbable, true)

    return spec.keyOwner, userId
end

function MultiplayerVehicleKey:getCurrentKeyOwnerId()
    if self.spec_enterableMultiplayerKey.keyOwner ~= nil then
        return self.spec_enterableMultiplayerKey.keyOwner.id
    end

    return -1
end

function MultiplayerVehicleKey:getUserCurrentVehicleKey(userId)
    if userId == nil then
        return g_currentMission.currentVehicleKey
    end

    if g_multiplayerVehicleKeys ~= nil then
        return g_multiplayerVehicleKeys[userId]
    end

    return nil
end

function MultiplayerVehicleKey:getIsKeyOwnerFromUniqueUserId(uniqueUserId)
    if self.spec_enterableMultiplayerKey.keyOwner ~= nil then
        return self.spec_enterableMultiplayerKey.keyOwner.uniqueUserId == uniqueUserId
    end

    return false
end

function MultiplayerVehicleKey:canMasterUserAccess()
    return self.spec_enterableMultiplayerKey.masterUserCanAccess
end

function MultiplayerVehicleKey:getKeyOwnerName()
    return self.spec_enterableMultiplayerKey.keyOwnerName or ""
end

function MultiplayerVehicleKey:getHasKeyOwner()
    return self.spec_enterableMultiplayerKey.keyOwner ~= nil
end

function MultiplayerVehicleKey.getCurrentUserFromPlayerStyle(playerStyle)
    if playerStyle ~= nil then
        for _, user in ipairs(g_currentMission.userManager.users) do
            if user.nickname == playerStyle.playerName then
                return user
            end
        end
    end

    return nil
end

function MultiplayerVehicleKey.getActionEventText(spec)
    if spec.keyOwner ~= nil then
        return spec.actionReturnKey
    end

    return spec.actionRemoveKey
end

function MultiplayerVehicleKey.addSideNotification(isOnline, text, vehicleName)
    if isOnline and g_currentMission ~= nil and g_currentMission.gameStarted then
        g_currentMission.hud:addSideNotification(MultiplayerVehicleKey.INGAME_NOTIFICATION_KEYS, string.format(text, tostring(vehicleName)), MultiplayerVehicleKey.NOTIFICATION_TIME)
    end
end

-- Needs to be done first so all specs work so 'onEnterVehicle' is no good here.
function MultiplayerVehicleKey:enterVehicle(superFunc, isControlling, playerStyle, farmId)
    self.spec_enterableMultiplayerKey.currentUser = MultiplayerVehicleKey.getCurrentUserFromPlayerStyle(playerStyle)
    superFunc(self, isControlling, playerStyle, farmId)
end

function MultiplayerVehicleKey:getCanMotorRun(superFunc)
    if self:getIsVehicleKeyAvailable(false) then
        return superFunc(self)
    end

    return false
end

function MultiplayerVehicleKey:getMotorNotAllowedWarning(superFunc)
    if self:getIsVehicleKeyAvailable(false) then
        return superFunc(self)
    end

    return string.format(self.spec_enterableMultiplayerKey.invalidUserWarning, self:getKeyOwnerName())
end

function MultiplayerVehicleKey:setIsTabbable(superFunc, isTabbable, ignoreMultiplayerVehicleKey)
    superFunc(self, isTabbable)

    if not ignoreMultiplayerVehicleKey then
        self.spec_enterableMultiplayerKey.isTabbable = isTabbable
    end
end
